#include "driver.h"
#include "layout.h"
#include "route.h"
#include "trackiter.h"
#include "train.h"
#include "trackchain.h"
#include "trainroutemetric.h"
#include "trainrouteplanner.h"
#include "trainrouter.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

TrainRouter::TrainRouter(Train &t):
	TrainAI(t),
	priority(0),
	arrival(ON_THE_WAY),
	waypoints_changed(false),
	metrics_stale(false),
	current_sequence(0),
	sequence_check_pending(false)
{
	train.get_layout().signal_block_reserved.connect(sigc::mem_fun(this, &TrainRouter::block_reserved));
	train.signal_advanced.connect(sigc::mem_fun(this, &TrainRouter::train_advanced));
	train.signal_rear_advanced.connect(sigc::mem_fun(this, &TrainRouter::train_rear_advanced));
}

TrainRouter::~TrainRouter()
{
	for(vector<TrainRouteMetric *>::iterator i=metrics.begin(); i!=metrics.end(); ++i)
		delete *i;
}

void TrainRouter::set_priority(int p)
{
	priority = p;
}

bool TrainRouter::set_route(const Route *r)
{
	routes.clear();
	if(r)
	{
		routes.push_back(r);
		create_lead_route();
	}

	waypoints.clear();
	sequence_points.clear();
	current_sequence = 0;
	sequence_check_pending = false;

	route_changed();

	return true;
}

const Route *TrainRouter::get_route() const
{
	if(routes.empty())
		return 0;
	return routes.front();
}

void TrainRouter::use_planned_route()
{
	if(!planner || planner->get_result()!=TrainRoutePlanner::COMPLETE)
		return;
	if(waypoints.empty())
		return;

	const list<Route *> &planned_routes = planner->get_routes_for(train);

	routes.clear();
	routes.insert(routes.end(), planned_routes.begin(), planned_routes.end());
	create_lead_route();

	sequence_points = planner->get_sequence_for(train);
	current_sequence = 0;
	sequence_check_pending = false;

	route_changed();
}

void TrainRouter::route_changed()
{
	BlockIter fncb = train.get_last_critical_block().next();

	arrival = ON_THE_WAY;
	reserving_route = routes.begin();
	if(!routes.empty())
	{
		/* Find the route that should be used for the next allocated block.  We
		can't rely on the resync code in block_reserved since we may need to
		clear the stop marker to continue allocation. */
		const BlockAllocator &allocator = train.get_block_allocator();
		TrackIter track = allocator.first().track_iter();
		list<SequencePoint>::iterator seq_begin = sequence_points.begin();
		for(; track; track=track.next())
		{
			if(!advance_to_track(reserving_route, track))
			{
				arrival = (allocator.is_block_current(track->get_block()) ? ADVANCED_TO_END : RESERVED_TO_END);
				break;
			}
			if(&track->get_block()==fncb.block())
				break;

			if(seq_begin!=sequence_points.end() && seq_begin->block==&track->get_block())
			{
				// Assume any sequence points within critical blocks to be cleared
				current_sequence = seq_begin->sequence_out;
				++seq_begin;
			}
		}

		sequence_points.erase(sequence_points.begin(), seq_begin);

		if(!sequence_points.empty())
		{
			const SequencePoint &sp = sequence_points.front();
			if(sp.block==fncb.block() && sp.preceding_train)
			{
				arrival = WAITING_FOR_SEQUENCE;
				sequence_check_pending = true;
			}
		}
	}

	/* Refresh from the first non-critical block to pick up any changes in the
	route.  Set stop marker first in case an arrival condition was met within the
	critical blocks. */
	if(arrival)
		train.stop_at(&*fncb.flip());
	train.refresh_blocks_from(*fncb);
	// If no arrival condition was found, clear a possible previous stop marker.
	if(!arrival)
		train.stop_at(0);

	const Route *route = get_route();
	signal_route_changed.emit(route);
	signal_event.emit(Message("route-changed", route));
}

void TrainRouter::set_destination(const TrackChain &d)
{
	if(waypoints.empty())
		waypoints.push_back(&d);
	else
		waypoints.back() = &d;
	waypoints_changed = true;
	metrics_stale = true;
}

void TrainRouter::add_waypoint(const TrackChain &wp)
{
	waypoints.push_back(&wp);
	waypoints_changed = true;
	metrics_stale = true;
}

const TrackChain &TrainRouter::get_waypoint(unsigned index) const
{
	if(index>=waypoints.size())
		throw out_of_range("TrainRouter::is_waypoint");

	return *waypoints[index];
}

const TrainRouteMetric &TrainRouter::get_metric(int index) const
{
	if(waypoints.empty())
		throw logic_error("no metrics");
	else if(metrics_stale)
		throw logic_error("metrics are stale");

	if(index<0)
		return *metrics.back();
	else if(static_cast<unsigned>(index)>=waypoints.size())
		throw out_of_range("TrainRouter::get_metric");
	else
		return *metrics[index];
}

void TrainRouter::set_departure_delay(const Time::TimeDelta &d)
{
	delay = d;
	waypoints_changed = true;
}

void TrainRouter::set_trip_duration(const Time::TimeDelta &d)
{
	duration = d;
}

void TrainRouter::message(const Message &msg)
{
	if(msg.type=="set-route")
	{
		if(msg.value.check_type<Route *>())
			set_route(msg.value.value<Route *>());
		else
			set_route(msg.value.value<const Route *>());
	}
	else if(msg.type=="clear-route")
		set_route(0);
	else if(msg.type=="set-destination")
	{
		if(msg.value.check_type<TrackChain *>())
			set_destination(*msg.value.value<TrackChain *>());
		else
			set_destination(*msg.value.value<const TrackChain *>());
	}
	else if(msg.type=="add-waypoint")
	{
		if(msg.value.check_type<TrackChain *>())
			add_waypoint(*msg.value.value<TrackChain *>());
		else
			add_waypoint(*msg.value.value<const TrackChain *>());
	}
	else if(msg.type=="set-departure-delay")
		set_departure_delay(msg.value.value<Time::TimeDelta>());
	else if(msg.type=="set-trip-duration")
		set_trip_duration(msg.value.value<Time::TimeDelta>());
}

void TrainRouter::tick(const Time::TimeDelta &dt)
{
	if(waypoints_changed && !planner)
		start_planning(train.get_layout());

	if(planner && planner->check()!=TrainRoutePlanner::PENDING)
		apply_plan(train.get_layout(), *planner);

	Layout &layout = train.get_layout();
	if(!layout.get_driver().is_halted() && !layout.get_clock().is_stopped())
	{
		if(delay)
		{
			delay -= dt;
			if(delay<Time::zero)
			{
				duration = max(duration+delay, Time::zero);
				delay = Time::zero;
			}
		}
		else if(duration)
			duration = max(duration-dt, Time::zero);
	}

	if(sequence_check_pending)
	{
		if(sequence_points.front().is_cleared())
		{
			arrival = ON_THE_WAY;
			train.stop_at(0);
		}
		sequence_check_pending = false;
	}

	if(arrival==ADVANCED_TO_END && !train.get_speed())
	{
		signal_arrived.emit(waypoints.back());
		signal_event.emit(Message("arrived", waypoints.back()));
		arrival = ARRIVED;
	}
	else if(arrival==ARRIVED && !train.get_block_allocator().is_active())
		set_route(0);
}

void TrainRouter::save(list<DataFile::Statement> &st) const
{
	st.push_back((DataFile::Statement("priority"), priority));

	if(!routes.empty())
	{
		RouteList::const_iterator i = routes.begin();
		for(; (i!=routes.end() && (*i)->is_temporary()); ++i) ;
		if(i!=routes.end())
			st.push_back((DataFile::Statement("route"), (*i)->get_name()));
	}
}

void TrainRouter::block_reserved(Block &block, Train *t)
{
	if(routes.empty())
		return;

	if(t!=&train)
	{
		if(!t)
			return;

		// Are we waiting for the other train to pass a sequence point?
		SequencePoint &sp = sequence_points.front();
		if(sp.preceding_train==t && sp.block==&block)
			/* The other train's router will advance its sequence on the same
			signal and may not have handled it yet. */
			sequence_check_pending = true;

		return;
	}

	// Did we reach our next sequence point?
	if(!sequence_points.empty())
	{
		SequencePoint &sp = sequence_points.front();
		if(sp.block==&block)
		{
			current_sequence = sp.sequence_out;
			sequence_points.pop_front();
		}
	}

	TrackIter track = train.get_block_allocator().iter_for(block).track_iter();

	// Is the block a turnout?  If it is, set it to the correct path.
	if(unsigned taddr = block.get_turnout_address())
	{
		int path = (*reserving_route)->get_turnout(taddr);
		if(path>=0)
			track->set_active_path(path);
	}

	/* If the allocator has released blocks from the front, we may need to
	resync reserving_route. */
	if(reserving_route==routes.end() || !(*reserving_route)->has_track(*track))
	{
		reserving_route = routes.begin();
		arrival = ON_THE_WAY;
		track = t->get_block_allocator().first().track_iter();
		for(; track; track=track.next())
		{
			if(!advance_to_track(reserving_route, track))
				throw logic_error("internal error (reservation outside of route)");
			else if(&track->get_block()==&block)
				break;
		}
	}

	/* Keep reserving_route pointing to the route containing the block that is
	expected to be allocated next. */
	for(; track; track=track.next((*reserving_route)->get_path(*track)))
	{
		if(!advance_to_track(reserving_route, track))
		{
			// We've reached the end of the route.  Stop here.
			arrival = RESERVED_TO_END;
			train.stop_at(&block);
			return;
		}
		if(&track->get_block()!=&block)
			break;
	}

	// Do we need to wait for another train to pass?
	if(!sequence_points.empty())
	{
		SequencePoint &sp = sequence_points.front();
		if(sp.block==&track->get_block() && !sp.is_cleared())
		{
			arrival = WAITING_FOR_SEQUENCE;
			train.stop_at(&block);
		}
	}
}

void TrainRouter::train_advanced(Block &block)
{
	BlockIter b_iter = train.get_block_allocator().iter_for(block);

	if(!waypoints.empty())
	{
		// A waypoint is considered reached when the train has advanced through it.
		const TrackChain &wp = *waypoints.front();
		TrackIter t_iter = b_iter.track_iter();
		if(wp.has_track(*t_iter))
		{
			for(; t_iter; t_iter=t_iter.next())
			{
				if(!wp.has_track(*t_iter))
				{
					if(waypoints.size()==1)
					{
						if(arrival==RESERVED_TO_END)
							arrival = ADVANCED_TO_END;
					}
					else
					{
						waypoints.erase(waypoints.begin());
						metrics_stale = true;
						signal_waypoint_reached.emit(&wp);
						signal_event.emit(Message("waypoint-reached", &wp));
					}
					break;
				}
				else if(!block.has_track(*t_iter))
					break;
			}
		}
	}
}

void TrainRouter::train_rear_advanced(Block &block)
{
	Track &track = *train.get_block_allocator().iter_for(block).endpoint().track;

	// Drop any routes that are now completely behind the train.
	for(RouteList::iterator i=routes.begin(); i!=routes.end(); ++i)
		if((*i)->has_track(track))
		{
			if(i!=routes.begin())
			{
				routes.erase(routes.begin(), i);
				const Route *route = get_route();
				signal_route_changed.emit(route);
				signal_event.emit(Message("route-changed", route));
			}
			break;
		}
}

void TrainRouter::create_metrics()
{
	for(vector<TrainRouteMetric *>::iterator i=metrics.begin(); i!=metrics.end(); ++i)
		delete *i;
	metrics.clear();

	metrics_stale = false;

	if(waypoints.empty())
		return;

	for(vector<const TrackChain *>::const_iterator i=waypoints.begin(); i!=waypoints.end(); ++i)
		metrics.push_back(new TrainRouteMetric(**i));

	for(unsigned i=metrics.size()-1; i-->0; )
		metrics[i]->chain_to(*metrics[i+1]);
}

bool TrainRouter::create_lead_route()
{
	if(routes.empty() || !train.is_placed())
		return false;

	BlockIter lcb = train.get_last_critical_block();
	TrackIter last_track_rev = lcb.reverse().track_iter();

	unsigned count = 0;
	for(TrackIter i=last_track_rev; (i && i->get_block().get_train()==&train); i=i.next())
	{
		if(routes.front()->has_track(*i))
			++count;
		else if(count>0)
		{
			if(count==routes.front()->get_tracks().size())
				last_track_rev = i;
			break;
		}
	}

	TrackIter next_track = last_track_rev.flip();
	if(!routes.front()->has_track(*last_track_rev) && !routes.front()->has_track(*next_track))
	{
		Route *pf = Route::find(next_track, *routes.front());
		if(!pf)
			return false;

		routes.push_front(pf);
	}

	Route *lead = 0;
	for(TrackIter i=last_track_rev; (i && i->get_block().get_train()==&train); i=i.next())
	{
		if(!lead && !routes.front()->has_track(*i))
		{
			lead = new Route(train.get_layout());
			lead->set_name("Lead");
			lead->set_temporary(true);
			routes.push_front(lead);

			TrackIter j = i.flip();
			lead->add_track(*j);
			lead->add_track(*j.next());
		}

		if(lead)
			lead->add_track(*i);
	}

	return true;
}

bool TrainRouter::advance_to_track(RouteList::iterator &route, const TrackIter &track)
{
	Track &prev_track = *track.flip();
	unsigned taddr = (track->get_type().is_turnout() ? track->get_turnout_address() : 0);
	for(unsigned i=0; route!=routes.end(); ++i)
	{
		bool in_route = (*route)->has_track(*track);
		bool prev_in_route = (*route)->has_track(prev_track);
		bool known_path = (!taddr || (*route)->get_turnout(taddr)>=0);

		if(in_route && (known_path || !prev_in_route))
			return true;
		else if(i==0 || prev_in_route)
			++route;
		else
			throw logic_error("internal error (routes are not continuous)");
	}

	return false;
}

void TrainRouter::get_routers(Layout &layout, vector<TrainRouter *> &routers)
{
	const map<unsigned, Train *> &trains = layout.get_trains();
	routers.reserve(trains.size());
	for(map<unsigned, Train *>::const_iterator i=trains.begin(); i!=trains.end(); ++i)
		if(TrainRouter *router = i->second->get_ai_of_type<TrainRouter>())
			routers.push_back(router);
}

void TrainRouter::start_planning(Layout &layout)
{
	vector<TrainRouter *> routers;
	get_routers(layout, routers);

	for(vector<TrainRouter *>::const_iterator i=routers.begin(); i!=routers.end(); ++i)
		if((*i)->metrics_stale)
			(*i)->create_metrics();

	RefPtr<TrainRoutePlanner> planner = new TrainRoutePlanner(layout);
	for(vector<TrainRouter *>::const_iterator i=routers.begin(); i!=routers.end(); ++i)
	{
		(*i)->waypoints_changed = false;
		(*i)->planner = planner;
	}

	planner->plan_async();
}

void TrainRouter::apply_plan(Layout &layout, TrainRoutePlanner &planner)
{
	if(planner.get_result()==TrainRoutePlanner::FAILED)
		layout.emergency(0, "Route planning failed");

	vector<TrainRouter *> routers;
	get_routers(layout, routers);

	for(vector<TrainRouter *>::const_iterator i=routers.begin(); i!=routers.end(); ++i)
		if((*i)->planner.get()==&planner)
		{
			(*i)->use_planned_route();
			(*i)->planner = 0;
		}
}


TrainRouter::SequencePoint::SequencePoint(Block &b, unsigned o):
	block(&b),
	preceding_train(0),
	sequence_in(0),
	sequence_out(o)
{ }

bool TrainRouter::SequencePoint::is_cleared() const
{
	if(!preceding_train)
		return true;

	TrainRouter *router = preceding_train->get_ai_of_type<TrainRouter>();
	return router->get_current_sequence()>=sequence_in;
}


TrainRouter::Loader::Loader(TrainRouter &r):
	DataFile::ObjectLoader<TrainRouter>(r)
{
	add("priority", &TrainRouter::priority);
	add("route",    &Loader::route);
}

void TrainRouter::Loader::route(const string &n)
{
	obj.set_route(&obj.train.get_layout().get_route(n));
}

} // namespace R2C2
